# -*- coding: utf-8 -*-
"""
Created on Sat Sep 28 22:03:07 2024

@author: richie bao
"""
import mesa
import numpy as np
import math


class Movement:

    @staticmethod
    def rt(heading, angle):
        """Turn right by the given angle (degrees)."""
        return (heading + angle) % 360

    @staticmethod
    def lt(heading, angle):
        """Turn right by the given angle (degrees)."""
        return (heading - angle) % 360

    @staticmethod
    def fd(agent, pos, heading, model, distance=1, torus=True):
        """Move forward by the given distance in the direction of the heading."""
        # Convert the heading to radians for trigonometric calculations
        heading_rad = math.radians(heading)

        # Calculate the new position based on the heading
        new_x = pos[0] + distance * math.cos(heading_rad)
        new_y = pos[1] + distance * math.sin(heading_rad)

        # Convert new position to integer grid position
        new_pos = (round(new_x), round(new_y))

        # Ensure the new position is within the grid bounds
        if torus:
            model.grid.move_agent(agent, new_pos)
        else:
            if model.grid.out_of_bounds(new_pos):
                print(
                    f"New position {new_pos} is out of bounds. Movement ignored."
                )
            else:
                # Move the agent to the new position
                # print("+++",new_pos)
                model.grid.move_agent(agent, new_pos)


class Person(mesa.Agent):

    def __init__(self, unique_id, pos, model, health=5, heading=0):
        super().__init__(unique_id, model)
        self.pos = pos
        self.heading = heading
        self.health = health

    def step(self):
        self.wander()
        self.reproduce()
        self.maybe_plant()
        self.eat_pollution()
        self.maybe_die()

    def wander(self):
        # neighbors = self.model.grid.get_neighborhood(self.pos, moore=True, include_center=False)
        # self.model.grid.move_agent_to_one_of(self,neighbors)
        self.heading = Movement.rt(self.heading, self.random.uniform(0, 50))
        self.heading = Movement.lt(self.heading, self.random.uniform(0, 50))
        Movement.fd(self, self.pos, self.heading, self.model, distance=1)
        self.health -= 0.1

    def reproduce(self):
        if self.health > 4 and self.random.random() < self.model.birth_rate:
            new_person = Person(
                self.model.next_id(),
                self.pos,
                self.model,
                5,
                self.random.uniform(0, 360),
            )
            self.model.grid.place_agent(new_person, self.pos)
            self.model.schedule.add(new_person)

    def maybe_plant(self):
        if self.random.random() < self.model.planting_rate:
            new_tree = Tree(
                self.model.next_id(), self.pos, self.model, health=5
            )
            self.model.grid.place_agent(new_tree, self.pos)
            self.model.schedule.add(new_tree)

    def eat_pollution(self):
        cell_content = self.model.grid.get_cell_list_contents([self.pos])
        patch = next(filter(lambda x: isinstance(x, Patch), cell_content), None)
        if patch and patch.pollution > 0.5:
            self.health -= patch.pollution / 10

    def maybe_die(self):
        if self.health <= 0:
            self.model.grid.remove_agent(self)
            self.model.schedule.remove(self)


class Tree(mesa.Agent):

    def __init__(self, unique_id, pos, model, health=5):
        super().__init__(unique_id, model)
        self.pos = pos
        self.health = health

    def step(self):
        self.cleanup()
        self.maybe_die()

    def cleanup(self):
        cell_content = self.model.grid.get_cell_list_contents([self.pos])
        patch = next(filter(lambda x: isinstance(x, Patch), cell_content), None)
        if patch:
            patch.pollution = max(0, patch.pollution - 1)
            self.health -= 0.1
            for neighbor in self.model.grid.get_neighbors(self.pos, moore=True):
                if isinstance(neighbor, Patch):
                    neighbor.pollution = max(0, neighbor.pollution - 0.5)

    def maybe_die(self):
        if self.health <= 0:
            self.model.grid.remove_agent(self)
            self.model.schedule.remove(self)


class Patch(mesa.Agent):

    def __init__(
        self, unique_id, pos, model, pollution=0, is_power_plant=False
    ):
        super().__init__(unique_id, model)
        self.pos = pos
        self._row = pos[0]
        self._col = pos[1]

        self.pollution = 0
        self.is_power_plant = False

    def step(self):
        self.pollute_diffusion()
        # if self.pollution>0:print("+++",self.pollution)
        # print(self.pollution)

    def pollute_diffusion(self):
        if self.is_power_plant:
            neighbors = self.model.grid.get_neighborhood(
                self.pos, moore=True, include_center=False
            )
            for neighbor in neighbors:
                this_cell = self.model.grid.get_cell_list_contents([neighbor])
                for obj in this_cell:
                    if isinstance(obj, Patch):
                        obj.pollution += (
                            self.pollution * 0.8 / 8
                        )  # Diffusion rate
                        # self.pollution -= self.pollution * 0.8  # Reduce current patch value

    def get_col(self):
        """Return the col location of this cell."""
        return self._col

    def get_row(self):
        """Return the row location of this cell."""
        return self._row

    def get_state(self):  # is_powerPlant
        """Return the current state (OPINION) of this cell."""
        return self.is_power_plant
        # return self.pollution
