# -*- coding: utf-8 -*-
"""
Created on Sat Sep 28 22:03:42 2024

@author: richie bao
"""
import mesa
from scheduler import RandomActivationByTypeFiltered
from agents import Person, Tree, Patch
import numpy as np

import warnings

warnings.simplefilter("ignore", category=Warning)


class PollutionModel(mesa.Model):
    height = 50
    width = 50
    initial_population = 30
    birth_rate = 0.1
    planting_rate = 0.05
    power_plants = 2
    polluting_rate = 3

    def __init__(
        self,
        height=50,
        width=50,
        initial_population=30,
        birth_rate=0.1,
        planting_rate=0.05,
        power_plants=2,
        polluting_rate=3,
    ):
        super().__init__()
        self.height = height
        self.width = width

        self.polluting_rate = polluting_rate
        self.birth_rate = birth_rate
        self.planting_rate = planting_rate
        self.initial_population = initial_population
        self.power_plants = power_plants

        self.grid = mesa.space.MultiGrid(self.width, self.height, torus=True)
        self.schedule = RandomActivationByTypeFiltered(self)

        self.datacollector = mesa.DataCollector(
            {
                "Person": lambda m: m.schedule.get_type_count(Person),
                "Tree": lambda m: m.schedule.get_type_count(Tree),
                "PollutionSum": lambda m: sum(
                    [
                        a.pollution
                        for a in m.schedule.agents_by_type[Patch].values()
                    ]
                ),
            }
        )

        for agent, (x, y) in self.grid.coord_iter():
            patch = Patch(self.next_id(), (x, y), self)
            self.grid.place_agent(patch, (x, y))
            self.schedule.add(patch)

        for _ in range(self.initial_population):
            x, y = self.random_xy()
            person = Person(
                self.next_id(), (x, y), self, 5, self.random.uniform(0, 360)
            )
            self.grid.place_agent(person, (x, y))
            self.schedule.add(person)

        self.create_power_plants(self.power_plants)

        self.running = True
        self.datacollector.collect(self)

    def create_power_plants(self, power_plants):
        xy_lst = []
        for _ in range(power_plants):
            xy_lst.append(self.random_xy())

        cells = self.grid.get_cell_list_contents(xy_lst)
        patches = [patch for patch in cells if isinstance(patch, Patch)]
        for p in patches:
            p.is_power_plant = True
            p.pollution = self.polluting_rate

    def random_xy(self):
        x = self.random.randrange(self.grid.width)
        y = self.random.randrange(self.grid.height)

        return (x, y)

    def step(self):
        self.schedule.step()
        self.datacollector.collect(self)
